import ee
import geopandas as gpd
import json
import numpy as np
import pandas as pd
from pathlib import Path
import s3fs as s3
from shapely import Point, LineString
from shapely.ops import split
import xarray as xr

def boxMethod(box, ls, width):
    '''
    apply box method for calculating relative terminus position
    inputs:
        box: shapely polygon
        ls: shapely line string
        width: width of box
    outputs: width averaged distance from 'baseline' of box to ls
    assumes/demands - that 'baseline' is to RHS of ls.
    '''
    splitted = split(box,ls)
    if len(splitted.geoms) == 1:
        baseline = np.nan
    else:
        baseline = splitted.geoms[1].area / width
    return baseline

def errWgtAvg(val, err, freq='AS'):
    '''
    compute error weighted average, and resample to
    input:
        val - the value to average
        err - the error associated with each val
        freq - resampled frequency
        resmpName, resamperrName - strings  for renaming output variables
    ALL INPUTS MUST BE xr.dataarrays. with mid_date dimension labelled as 'mid_date'.
    output:
        weighted_avg: xarray dataset of error weighted average resampled to desired frequency
        combined_uncert: resultant uncertainty
    '''
    empty = np.zeros(val.shape)
    empty[:] = np.nan
    err_masked = xr.where(np.isnan(val), empty, err) # takes account of potential nans in vals that aren't present weights
    weights = 1 / (err_masked**2)
    top = (weights * val).resample(mid_date = freq).sum()
    bottom = weights.resample(mid_date = freq).sum()
    weighted_avg = top/bottom
    combined_uncert = (np.sqrt(1 / (weights.resample(mid_date = freq).sum())))
    stdDev = val.resample(mid_date = freq).std()
    return weighted_avg.rename('v'), combined_uncert.rename('err'), stdDev.rename('std')

def v_uncert(vx,vy,vxerr=0,vyerr=0):
    '''
    returns resultant velocity (v), with uncertainty (err) 
    from vx, vy components
    each with their uncertainty vxerr, and vyerr
    where v = sqrt(vx^2 + vy^2)
    operates at single point
    '''
    def pow_err(q,x,dx,n): # where q = val^power
        '''
        returns uncertainty in q (dq), where q = x^n, and x has uncertainty dx
        dq / |q| = |n|* dx / |x|
        '''
        return np.abs(q) * np.abs(n) * (dx / np.abs(x))
    vx2err = pow_err(vx**2, vx, vxerr, 2) # error in vx^2
    vy2err = pow_err(vy**2, vy, vyerr, 2) # error in vy^2    
    sq_sum = vx**2 + vy**2
    sq_sum_err = np.hypot(vx2err, vy2err) # error in vx^2 + vy^2    
    v = np.sqrt(sq_sum) # resultant velocity    
    err = pow_err(v, sq_sum, sq_sum_err, 0.5) # error in resultant velocity    
    return v, err

def getCubeInstances(ps):
    '''
    get itsLive cube outlines
    intersect with points that you want to sample at
    identify cube that each point falls into
    add cube url to sample point dataframe
    get list of unique cubes
    inputs - ps (samplePointS) - geopandas dataframe, geometry points
    outputs - zarrlist: list of zarrcubes
            - updated ps with zarrcubeurls for each point
    '''
    access = s3.S3FileSystem(anon = True)
    # get outlines of cubes
    with access.open("s3://its-live-data/datacubes/catalog_v02.json", 'r') as something:
        allOutlines = gpd.GeoDataFrame.from_features(json.load(something), crs = 4326)

    # identify which cubes points fall into 
    ps = (ps.to_crs(4326).overlay(allOutlines[['geometry','zarr_url']],
                                how = 'intersection')
        .set_index(['glacier','cd'])
        .to_crs(3413))

    # update url 
    ps['zarr_url'] = ps['zarr_url'].apply(lambda q: (q.replace("http:", "s3:")
                                                    .replace(".s3.amazonaws.com", "")))

    # list of unique cubes
    zarrlist = ps['zarr_url'].unique().tolist()
    print('number of unique zarr cubes:', len(zarrlist))
    return zarrlist, ps

def sampleCubes(zarrlist, ps, spatialAveraging=True):
    '''
    sampling each zarr cube in `zarrlist`, and each point in `ps`
    if spatialAveraging==True sample in a 3x3 window around each point
    inputs: zarrlist: list of zarrcubes
            ps: geopandas dataframe of points to be sampled
            sptialAveraging: True/False
    returns: return `dss` is of ITSLIVE xarray.Datasets that have been sampled, so their dimemsions are:
        `mid_date`, `glacier`, `cd` (and if spatialAveraging==True, `k`)
        where `cd` is binned distance along centreline, and `k` is integer 0-8 representing the 9 points in a 3x3 window
    '''
    
    if spatialAveraging: ps['windows'] = pd.Series(dtype = object)

    dss = [] # empty list to fill with each sampled xr.Dataset
    for z in zarrlist:
        # index of sample points that are within the current zarr
        zidx = ps['zarr_url'] == z
        
        with xr.open_dataset(z, engine = "zarr", storage_options = {'anon':True}) as cube:        
            
            if spatialAveraging: # need to generate additional points to sample around each vertex 
                # get grid closest grid square indices
                diffsX = np.abs(ps.loc[zidx, 'x'].values - cube['x'].values[:,np.newaxis])
                diffsY = np.abs(ps.loc[zidx, 'y'].values - cube['y'].values[:,np.newaxis])
                idxX = diffsX.argmin(axis = 0)
                idxY = diffsY.argmin(axis = 0)

                # construct 3x3 kernal about each index
                windows = []
                for X,Y in list(zip(idxX,idxY)):
                    wind = []
                    for y in [Y+1, Y, Y-1]:
                        for x in [X-1, X, X+1]:
                            wind.append((x,y)) # tuple of grid indices
                    windows.append(wind)
                
                # add windows indicies into smaple point data frame
                ps.loc[zidx, 'windows'] = pd.Series(windows,
                                                    index=ps.loc[zidx].index)

                # unpack list of windows, set index to glacier, cumulative distance, and 'k'
                # where 'k' denotes each gridsquare within the 3x3 window (0-8)
                queryP = (ps.loc[zidx, 'windows'].apply(pd.Series)
                          .reset_index()
                          .melt(['glacier','cd'])
                          .sort_values(by = ['glacier', 'cd'])
                          .rename(columns = {'variable':'k','value':'idx'})
                          .set_index(['glacier','cd','k']))

                # unpack each window index tuple
                queryP = (queryP['idx'].apply(pd.Series)
                          .rename(columns = {0:'xidx',1:'yidx'}))
                
                # remove invalid indicies occurs when trying to add points that cross cube boundary....
                invalidQ = ((queryP['yidx'] >= len(cube['y'])) 
                            | (queryP['yidx'] < 0)
                            | (queryP['xidx'] >= len(cube['x']))
                            | (queryP['xidx'] < 0))
                queryP = queryP[~invalidQ]
                # print(np.count_nonzero(invalidQ))
                
                # create xarrays for querying zarrcube
                queryX = queryP['xidx'].to_xarray()
                mask = ~np.isnan(queryX)         # some nan's when dealing with centrelines of differing lengths   
                queryX = xr.where(mask, queryX, 0).astype(int) # have to be 'ints'
                        
                queryY = queryP['yidx'].to_xarray()
                queryY = xr.where(mask, queryY, 0).astype(int)
                
                # query cube *and rename*
                cube = (cube.isel(x = queryX,
                                  y = queryY)
                        .sortby('mid_date'))
                
            else: # if not spatially averaging, rely on .sel(method = 'nearest')
                # create 
                queryX = ps.loc[zidx, 'x'].to_xarray()
                queryY = ps.loc[zidx, 'y'].to_xarray()
                mask = ~np.isnan(queryX)
                
                cube = (cube.sel(x = queryX,
                                y = queryY,
                                method = 'nearest')
                        .sortby('mid_date'))
            
            cube['x'] = xr.where(mask, cube['x'], np.nan)
            cube['y'] = xr.where(mask, cube['y'], np.nan)
            cube['mask'] = mask
            
            print(cube.dims)        
            dss.append(cube.chunk(chunks = {'cd':15,'glacier':1}))

    print(len(dss), 'cubes loaded\n and chunked')
           
    # NOTE when multiple cubes loaded - cannot simply use xr.concat([<list of cubes>])
    # as mid_dates, are unique/duplicates
    
    return dss #, ps
    
def getAnnualAverages(dss, ddt_min, ddt_max):
    '''
    calculate annual average velocities along centrelines
    inputs: list of sampled itslive cubes (dims: mid_date, glacier, cd, (maybe)k)
            infer if cube needs to be spatially averaged if `k` is present in dims
            ddt_min / dd_max: pandas dateoffset strings e.g. '430D'. for filtering itslive cube to only include 
                velocities derived from image pairs with date_dt values between ddt_min and ddt_max
    returns:    df: pandas dataframe of annual average velocities,
        and their errors for each glacier and each point along centreline
                allV: pandas dataframe of spatially averaged velocity measurements at ALL timesteps in original xr.Datasets
        this is useful for getting standard deviations/CIs when using seaborn plotting functions
        requires using .groupby(['glacier','cd','year])
    '''
    annualVs = []
    annualErrs = []
    annualStds = []
    for ds in dss:
        ddt_idx = (ds['date_dt'] >= pd.Timedelta(ddt_min)) & (ds['date_dt'] <= pd.Timedelta(ddt_max)) # filter date_dt
        if 'k' in list(ds.dims): #spatialAveraging:             
                # compute average velocity in each kernal, from vx,vy components
                avg_vx = ds['vx'][ddt_idx,:,:,:].mean(dim = 'k', skipna = True)
                avg_vy = ds['vy'][ddt_idx,:,:,:].mean(dim = 'k', skipna = True)
                avg_v, avg_v_err = v_uncert(avg_vx, avg_vy,
                                            ds['vx_error'][ddt_idx],
                                            ds['vy_error'][ddt_idx])
                # or just do ds['v'].mean(dim = 'k')...and use the supplied 'v_error'
                annualV, annualErr, annualStd = errWgtAvg(avg_v, avg_v_err)
                
                output_shape = dict(zip(annualV.dims, annualV.shape))
                valid_idx = np.repeat((ds['mask'].mean(dim = 'k').values==1)[np.newaxis,:,:],
                                repeats = output_shape['mid_date'],
                                axis = 0)
                
                annualV = xr.where(valid_idx, annualV, np.nan)
                annualErr = xr.where(valid_idx, annualErr, np.nan)
                annualStd = xr.where(valid_idx, annualStd, np.nan)
                
                annualVs.append(annualV)
                annualErrs.append(annualErr)
                annualStds.append(annualStd)
        
        else:
                annualV, annualErr, annualStd = errWgtAvg(ds['v'][ddt_idx,:,:],
                                                          ds['v_error'][ddt_idx,:,:])
                annualVs.append(annualV)
                annualErrs.append(annualErr)
                
    v = xr.merge(annualVs)['v']
    err = xr.merge(annualErrs)['err']
    # stdDev = xr.merge(annualStds)['std']

    # convert to data frame to get better control over plotting
    df = v.to_dataframe(name = 'v', dim_order = ['glacier','cd','mid_date'])
    df = df.merge(err.to_dataframe(name = 'err', dim_order = ['glacier','cd','mid_date']), left_index = True, right_index = True)
    # df = df.merge(stdDev.to_dataframe(name = 'std', dim_order = ['glacier','cd','mid_date']), left_index = True, right_index = True)
    
    df['year'] = df.index.get_level_values(level=2).year
    df['pcnt'] = (df['v'] / df.loc[df.groupby(['glacier', 'cd'])
                ['year'].transform('idxmin'), 'v'].values) * 100
    df.reset_index(inplace = True)

    allV = avg_v.to_dataframe(name = 'v', dim_order = ['glacier','cd','mid_date'])
    allV['year'] = allV.index.get_level_values(level=2).year
    allV.reset_index(inplace = True)
    
    return df, allV

def getSeasonalAverages(dss, startD, stopD, ddt_max='30D',  rsmpFreq='Q-FEB'):
    '''
    get seasonally averaged velocities
    input:  dss - list of xr.Datasets dims = [mid_date, glacier, cd, (k)]
            startD, stopD - pandas datetimes for filtering on mid_date
            ddt_max - pandas dateoffset string for filtering date_dt - default 30 days
            rsmpFreq - frequnecy to resample to, default = Q-FEB
    returns: sv_df : a pandas data frame with columns:
                glacier, cd, mid_date (with frequency == rsmpFreq), v, season, year
    '''
    
    seasonalVs = []
    # seasonalErrs = []
    # seasonalStds = []
    for ds in dss:
        
        # date range
        d_idx = (ds['mid_date'].to_pandas().between(startD, stopD))
        # date_dt filter
        ddt_idx = ds['date_dt'] <= pd.Timedelta(ddt_max)
        # combined mid_date filter
        mid_date_idx = ddt_idx & d_idx
        
        if 'k' in list(ds.dims): #spatialAveraging: if spatialAveraging
                # compute average velocity in each kernal, from vx,vy components
                avg_vx = ds['vx'][mid_date_idx,:,:,:].mean(dim = 'k', skipna = True)
                avg_vy = ds['vy'][mid_date_idx,:,:,:].mean(dim = 'k', skipna = True)
                avg_v, avg_v_err = v_uncert(avg_vx,
                                            avg_vy,
                                            ds['vx_error'][mid_date_idx],
                                            ds['vy_error'][mid_date_idx])
                # # or just do ds['v'].mean(dim = 'k')...and use the supplied 'v_error'
                seasonalV, _, _ = errWgtAvg(avg_v, avg_v_err, freq = rsmpFreq)
                
                output_shape = dict(zip(seasonalV.dims, seasonalV.shape))
                valid_idx = np.repeat((ds['mask'].mean(dim = 'k').values == 1)[np.newaxis,:,:],
                                repeats = output_shape['mid_date'],
                                axis = 0)
                
                seasonalV = xr.where(valid_idx, seasonalV, np.nan)
                # seasonalErr = xr.where(valid_idx, seasonalErr, np.nan)
                # seasonalStd = xr.where(valid_idx, seasonalStd, np.nan)
                
                seasonalVs.append(seasonalV)
                # seasonalErrs.append(seasonalErr)
                # seasonalStds.append(seasonalStd)
        
        else:
                avg_v = ds['v'][mid_date_idx,:,:]
                seasonalV, _, _ = errWgtAvg(ds['v'][mid_date_idx,:,:],
                                            ds['v_error'][mid_date_idx,:,:])
                # seasonalVs.append(seasonalV)
                # seasonalErrs.append(seasonalErr)
                seasonalVs.append(seasonalV)
                
    Sv = xr.merge(seasonalVs)['v']
    
    ## seasonal trends along centreline constructed by first calculating seasonal averages (1 per year) 
    # then computing trend
    ssn ={1:'DJF',2:'MAM',3:'JJA',4:'SON'}
    sv_df = Sv.to_dataframe(dim_order=['glacier','cd','mid_date'])
    sv_df.reset_index(inplace = True)
    sv_df['season'] = (sv_df['mid_date'].dt.month % 12 // 3 + 1).map(ssn)
    sv_df['year'] = sv_df['mid_date'].dt.year
    
    return sv_df

def getTimeseries(dss, startD, stopD, distances, ddt_max='30D'):
    '''
    for formatting timeseries at a point(s) along centreline, and resampling them to regular frequnecies
    inputs: dss - list of xarray datasets, dims(mid_date, glacier, cd, (k))
            startD, stopD - pandas datetimes for filtering - mid_date
            distances - list of distances along centreline (cd) to include
            ddt_max - pandas date offset string (e.g. '30D'. for filtering date_dt. only include date_dts <= ddt_max
            
    returns: ts_df:
        pandas dataframe: velocities that meet date, date_dt and cd filter requirements (spatially averaged if `k` is present in list of dims)
        w/  rolling 28 d average           
    '''
    filtered = []
    for ds in dss:        
        # filter on startD, stopD and ddt_max
        d_idx = (ds['mid_date'].to_pandas().between(startD, stopD))
        ddt_idx = ds['date_dt'] <= pd.Timedelta(ddt_max)
        mid_date_idx = ddt_idx & d_idx
                
        ## filter on distances
        cd_idx = ds['cd'].isin(distances)
        
        if 'k' in list(ds.dims): # then spatial averaging
            # average spatially in kernal.
            avg_vx = ds['vx'][mid_date_idx,:,cd_idx,:].mean(dim = 'k', skipna = True)
            avg_vy = ds['vy'][mid_date_idx,:,cd_idx,:].mean(dim = 'k', skipna = True)
            avg_v, avg_v_err = v_uncert(avg_vx, avg_vy,
                                        ds['vx_error'][mid_date_idx],
                                        ds['vy_error'][mid_date_idx])
                              
        else:
            avg_v, avg_v_err = v_uncert(ds['vx'][mid_date_idx,:,cd_idx],
                                        ds['vy'][mid_date_idx,:,cd_idx],
                                        ds['vx_error'][mid_date_idx],
                                        ds['vy_error'][mid_date_idx])
      
        # merge v, v_err and meta data
        filtered.append(xr.merge([avg_v.rename('v'),
                                avg_v_err.rename('err'),
                                ds[['acquisition_date_img1',
                                    'acquisition_date_img2',
                                    'date_dt',
                                    'satellite_img1',
                                    'mission_img1']].sel(mid_date = mid_date_idx)
                                ]))

    # merge list of filterd dataframes and convert to dask_df
    ds_out = xr.merge(filtered)
    ts_df = ds_out.to_dataframe()

    # # drop nan rows
    ts_df = ts_df[~ts_df['v'].isna()]
        
    # rolling median
    ts_df = ts_df.merge(ts_df
                        .reset_index(level = [1,2])
                        .groupby(['glacier','cd'])['v']
                        .rolling('28D')
                        .median()
                        .rename('smoothed_v')
                        .reset_index(),
                        left_on = ['mid_date','glacier','cd'],
                        right_on = ['mid_date','glacier','cd'])

    # calculate intermediate steps for error-weighted averages
    ts_df['v_over_sigma2'] = ts_df['v'] / (ts_df['err']**2)
    ts_df['one_over_sigma2'] = 1 / (ts_df['err']**2)
    
    return ts_df

def checkempty(img):
    '''
    for checking if there are any icesat offset corrections in ArcticDEM ee.Image
    adds 'isEmpty' field. True if no corrections.
    '''
    img = img.set('isEmpty', ee.Algorithms.IsEqual(img.get('registrationDX'),None))
    return img

def adjust(img):
    '''
    img, ee.Image
    uses dx,dy,dz corrections to translate image and add elevation offset
    returns ee.Image with new band `adjusted` containing the offset elevations
    '''
    _dx = img.get('registrationDX')
    _dy = img.get('registrationDY')
    _dz = img.get('registrationDZ')
    img = img.translate(_dx, _dy, units = 'meters')
    _zoff = ee.Image.constant(_dz).rename('dz')
    img = img.addBands((img.select('elevation')
                        .add(_zoff))
                       .rename('adjusted'))
    return img

def read_filter(path,
                ids,
                date_filter=pd.Timestamp('2010-12-31'),
                var='discharge'):
    '''
    helper for opening and filtering runoff .nc files
    path - file path to .nc file
    ids  list of catchments ids to select
    date_filter (pd.Timestamp) default 31st Dec 2010, only select times after this date
    var (str) name of variable in .nc files to select. default: discharge
    returns filtered, chunked, data array
    '''
    name = Path(path).stem
    with xr.open_dataset(path,
                         chunks={'time':-1,
                                 'station':1000}) as ds:
        ds = (ds.sel(
            station = ds['station'].isin(ids),  # select catchment ids
            time = ds['time'] >= date_filter    # filter on time
            )[var]
              .assign_coords(model=name)        # add new dimension for model name
              .expand_dims('model')
              )
        return ds